# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from sdaps import model
from sdaps import image
import shutil

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

def _insert_dummy_pages(survey, duplex_scan):
    # Insert dummy pages if the survey is duplex and the duplex option was not
    # passed
    if survey.defs.duplex:
        # One image per questionnaire page in duplex mode
        image_count_factor = 1
        # No dummy pages in duplex mode
        insert_dummy_pages = False
    else:
        # Two images per questionnaire page in duplex mode
        image_count_factor = 2

        # In simplex mode insertion of dummy pages depends on the command line
        # optoin (default is True)
        if duplex_scan:
            insert_dummy_pages = False
        else:
            insert_dummy_pages = True

    return insert_dummy_pages, image_count_factor

def check_image(survey, file, duplex_scan=False, force=False, message=False):

    insert_dummy_pages, image_count_factor = _insert_dummy_pages(survey, duplex_scan)

    if not image.check_tiff_monochrome(file):
        if message:
            print(_('Invalid input file %s. You need to specify a (multipage) monochrome TIFF as input.') % (file,))
        return False

    num_pages = image.get_tiff_page_count(file)

    c = survey.questionnaire.page_count
    if not insert_dummy_pages:
        c = c * image_count_factor

    # This test is on the image count that needs to come from the file
    if num_pages % c != 0 and not force:
        if message:
            print(_('Not adding %s because it has a wrong page count (needs to be a mulitple of %i).') % (file, c))
        return False

    return True

def add_image(survey, file, duplex_scan=False, force=False, copy=True):

    insert_dummy_pages, image_count_factor = _insert_dummy_pages(survey, duplex_scan)

    if not check_image(survey, file, duplex_scan, force, message=True):
        return

    num_pages = image.get_tiff_page_count(file)

    c = survey.questionnaire.page_count
    if not insert_dummy_pages:
        c = c * image_count_factor

    if insert_dummy_pages:
        c = c * image_count_factor

    if copy:
        tiff = survey.new_path('%i.tif')
        shutil.copyfile(file, tiff)
    else:
        tiff = file

    if copy:
        tiff = os.path.basename(tiff)
    else:
        tiff = os.path.relpath(os.path.abspath(tiff), survey.survey_dir)

    pages = list(range(num_pages))
    while len(pages) > 0:
        sheet = model.sheet.Sheet()
        survey.add_sheet(sheet)
        while len(pages) > 0 and len(sheet.images) < c:
            img = model.sheet.Image()
            sheet.add_image(img)
            img.filename = tiff
            img.tiff_page = pages.pop(0)

            # And a dummy page if required
            if insert_dummy_pages:
                img = model.sheet.Image()
                sheet.add_image(img)

                img.filename = "DUMMY"
                img.tiff_page = -1
                img.ignored = True




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

class QObject(object):

    def __init__(self, parent):
        self.review_comment = ''

    @property
    def empty(self):
        return not self.review_comment

class Box(object):

    def __init__(self, parent):
        self.state = 0
        self.metrics = dict()
        self.quality = 1
        self.x = parent.x
        self.y = parent.y
        self.width = parent.width
        self.height = parent.height

        # Set _parent last, so that we don't trigger notifications
        # during __init__
        self._parent = parent

    def __setattr__(self, name, value):
        if hasattr(self, name):
            old_value = getattr(self, name)
        else:
            old_value = None

        object.__setattr__(self, name, value)
        # private
        if name.startswith('_'):
            return

        if value != old_value and hasattr(self, '_parent') and self._parent is not None:
            self._dirty = True
            self._parent.question.questionnaire.notify_data_changed(self._parent, self, name, old_value)

    @property
    def empty(self):
        return not self.state

class Checkbox(Box):

    pass


class Textbox(Box):

    def __init__(self, parent):
        self.text = str()

        Box.__init__(self, parent)


class Additional_Mark(object):

    def __init__(self, parent):
        self.value = 0

    @property
    def empty(self):
        return self.value == 0




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008,2011, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import defs

from sdaps import model
from sdaps.utils.exceptions import RecognitionError

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

# Note, while it is possible to cache data in the image object, this data
# will not change when the image is switched to the next sheet. So one needs
# to be careful when to invalidate it.


# All of these functions can assume that the matrix has been recognized
# for the image.
# They may throw a RecognitionError if there was an unrecoverable error.
# If no data can be retrieved(because eg. it is not printed on that
# page) they may return None to indicate this.

class Image(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'style'
    obj_class = model.sheet.Image

    def get_page_rotation(self):
        # Returns page rotation or "None" if it cannot be retrieved

        res = get_pagenumber_and_rotation(self)
        if res is None:
            return None

        return res[0]

    def get_page_number(self):
        # Returns page number or "None" if it cannot be retrieved
        # Returns page number rotation or "None" if it cannot be retrieved

        res = get_pagenumber_and_rotation(self)
        if res is None:
            return None
        # The page may not be rotated at this point
        if res[0]:
            raise RecognitionError

        return res[1]

    def get_survey_id(self):
        # Returns the survey ID or "None" if it cannot be retrieved

        if self.obj.page_number % 2 == 0 or \
           self.obj.sheet.survey.questionnaire.page_count == 1:
            pos = self.obj.sheet.survey.defs.get_survey_id_pos()

            survey_id = read_codebox(self,
                                     pos[0], pos[2])
            survey_id = read_codebox(self,
                                     pos[1], pos[2],
                                     survey_id)

            return survey_id
        else:
            return None

    def get_questionnaire_id(self):
        # Returns the questionnaire ID or "None" if it cannot be retrieved

        if self.obj.page_number % 2 == 0 or \
           self.obj.sheet.survey.questionnaire.page_count == 1:
            pos = self.obj.sheet.survey.defs.get_questionnaire_id_pos()

            questionnaire_id = read_codebox(self,
                                            pos[0], pos[2])

            return questionnaire_id
        else:
            return None

    def get_global_id(self):
        # The classic style does not support a global ID property.
        return None


############################
# Internal Helpers
############################

def read_codebox(self, x, y, code=0):
    for i in range(defs.codebox_length):
        code <<= 1
        coverage = self.obj.recognize.get_coverage(
            x + (i * defs.codebox_step) + defs.codebox_offset,
            y + defs.codebox_offset,
            defs.codebox_step - 2 * defs.codebox_offset,
            defs.codebox_height - 2 * defs.codebox_offset
        )
        if coverage > defs.codebox_on_coverage:
            code += 1
    return code


def get_pagenumber_and_rotation(self):
    # The coordinates in defs are the center of the line, not the bounding box of the box ...
    # Its a bug im stamp
    # So we need to adjust them
    half_pt = 0.5 / 72.0 * 25.4
    pt = 1.0 / 72.0 * 25.4

    # Check whether there is a valid transformation matrix. If not simply return.
    if self.obj.matrix.mm_to_px(False) is None:
        return

    width = defs.corner_box_width
    height = defs.corner_box_height
    padding = defs.corner_box_padding
    survey = self.obj.sheet.survey
    corner_boxes_positions = [
        (defs.corner_mark_left + padding, defs.corner_mark_top + padding),
        (survey.defs.paper_width - defs.corner_mark_right - padding - width, defs.corner_mark_top + padding),
        (defs.corner_mark_left + padding, survey.defs.paper_height - defs.corner_mark_bottom - padding - height),
        (survey.defs.paper_width - defs.corner_mark_right - padding - width,
         survey.defs.paper_height - defs.corner_mark_bottom - padding - height)
    ]
    corners = [
        int(self.obj.recognize.get_coverage(
            corner[0] - half_pt,
            corner[1] - half_pt,
            width + pt,
            height + pt
        ) > defs.cornerbox_on_coverage)
        for corner in corner_boxes_positions
    ]

    try:
        page_number = defs.corner_boxes.index(corners) + 1
        rotated = False
    except ValueError:
        try:
            page_number = defs.corner_boxes.index(corners[::-1]) + 1
            rotated = True
        except ValueError:
            raise RecognitionError

    return rotated, page_number




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2012 Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import defs

from sdaps import model
from sdaps.utils.exceptions import RecognitionError
from sdaps.utils.barcode import read_barcode


# Reading the metainformation of CODE-128 style questionnaires. See classic.py
# for some more information.

class Image(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'style'
    obj_class = model.sheet.Image

    def get_page_rotation(self):
        # Returns page rotation or "None" if it cannot be retrieved

        # Figure out wether the page is rotated by looking for a barcode first
        # at the bottom right, then at the top left.
        # Note that we do not care about the value of the barcode, we are happy
        # to simply know that it exists.

        paper_width = self.obj.sheet.survey.defs.paper_width
        paper_height = self.obj.sheet.survey.defs.paper_height

        # Search for the barcode in the lower right corner.
        # Note that we cannot find another barcode this way, because the one in the
        # center of the page is not complete
        code = \
            read_barcode(self.obj.surface.surface, self.obj.matrix.mm_to_px(),
                         paper_width / 2,
                         paper_height - defs.corner_mark_bottom - defs.code128_vpad - defs.code128_height - 5,
                         paper_width / 2,
                         defs.corner_mark_bottom + defs.code128_vpad + defs.code128_height + 5)

        if code is None:
            # Well, that failed, so try to search the upper left corner instead
            code = \
                read_barcode(self.obj.surface.surface, self.obj.matrix.mm_to_px(),
                             0, 0,
                             paper_width / 2,
                             defs.corner_mark_bottom + defs.code128_vpad + defs.code128_height + 5)

            if code is not None:
                return True
            else:
                return None
        else:
            return False


    def get_page_number(self):
        # Returns page number or "None" if it cannot be retrieved

        # In this function assume that the rotation is correct already.
        paper_width = self.obj.sheet.survey.defs.paper_width
        paper_height = self.obj.sheet.survey.defs.paper_height

        # Search for the barcode in the lower right corner.
        code = \
            read_barcode(self.obj.surface.surface, self.obj.matrix.mm_to_px(),
                         paper_width / 2,
                         paper_height - defs.corner_mark_bottom - defs.code128_vpad - defs.code128_height - 5,
                         paper_width / 2,
                         defs.corner_mark_bottom + defs.code128_vpad + defs.code128_height + 5)

        # The code needs to be entirely numeric and at least 4 characters for the page
        if code is None or(not code.isdigit() and len(code) < 4):
            return None

        # The page number is in the lower four digits, simply extract it and convert
        # to integer
        return int(code[-4:])

    def get_survey_id(self):
        # Returns the survey ID or "None" if it cannot be retrieved

        # In this function assume that the rotation is correct already.
        paper_width = self.obj.sheet.survey.defs.paper_width
        paper_height = self.obj.sheet.survey.defs.paper_height

        # Search for the barcode in the lower left corner.
        code = \
            read_barcode(self.obj.surface.surface, self.obj.matrix.mm_to_px(),
                         paper_width / 2,
                         paper_height - defs.corner_mark_bottom - defs.code128_vpad - defs.code128_height - 5,
                         paper_width / 2,
                         defs.corner_mark_bottom + defs.code128_vpad + defs.code128_height + 5)

        if code is None or not code.isdigit() or len(code) <= 4:
            return None

        return int(code[:-4])

    def get_questionnaire_id(self):
        # Returns the questionnaire ID or "None" if it cannot be retrieved

        # In this function assume that the rotation is correct already.
        paper_width = self.obj.sheet.survey.defs.paper_width
        paper_height = self.obj.sheet.survey.defs.paper_height

        # Search for the barcode on the bottom left of the page
        code = \
            read_barcode(self.obj.surface.surface, self.obj.matrix.mm_to_px(),
                         0,
                         paper_height - defs.corner_mark_bottom - defs.code128_vpad - defs.code128_height - 5,
                         paper_width / 2,
                         defs.corner_mark_bottom + defs.code128_vpad + defs.code128_height + 5)

        # Simply return the code, it may be alphanumeric, we don't care here
        # XXX: Is that assumption sane?
        return code

    def get_global_id(self):
        # Returns the global ID or "None" if it cannot be retrieved

        # In this function assume that the rotation is correct already.
        paper_width = self.obj.sheet.survey.defs.paper_width
        paper_height = self.obj.sheet.survey.defs.paper_height

        # Search for the barcode in the bottom center of the page
        code = \
            read_barcode(self.obj.surface.surface, self.obj.matrix.mm_to_px(),
                         paper_width / 4,
                         paper_height - defs.corner_mark_bottom - defs.code128_vpad - defs.code128_height - 5,
                         paper_width / 2,
                         defs.corner_mark_bottom + defs.code128_vpad + defs.code128_height + 5)

        # Simply return the code, it may be alphanumeric, we don't care here
        return code



# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2013, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

_fallback = "A4", (210., 297.)

def _get_gtk_ppd_papersize(paper=None):
    try:
        import gi
        gi.require_version('Gtk', '3.0')
        from gi.repository import Gtk
    except:
        return paper, None
    else:
        def _find_papersize_by_ppd_name(paper=None):
            if not paper:
                return None

            paper = paper.lower()
            for papersize in Gtk.PaperSize.get_paper_sizes(False):
                if paper == papersize.get_ppd_name().lower():
                    return papersize

        papersize = _find_papersize_by_ppd_name(paper)
        if papersize is None:
            # Retrieve default paper size
            paper_name = Gtk.PaperSize.get_default()
            papersize = Gtk.PaperSize.new(paper_name)

        width = papersize.get_width(Gtk.Unit.MM)
        height = papersize.get_height(Gtk.Unit.MM)

        return papersize.get_ppd_name(), (width, height)

def get_tex_papersize(paper=None):
    # Assume that the name is the PPD name in lowercase + 'paper'
    paper, size = _get_gtk_ppd_papersize(paper)

    if paper is None:
        paper = _fallback[0]

    return paper.lower() + 'paper'

def get_reportlab_papersize(paper=None):
    paper, size = _get_gtk_ppd_papersize(paper)

    if size:
        size = (size[0] / 25.4 * 72.0, size[1] / 25.4 * 72.0)

    if size is None:
        if paper:
            from reportlab.lib import pagesizes
            if hasattr(pagesizes, paper.upper()):
                size = getattr(pagesizes, paper.upper())

    if size is None:
        size = (_fallback[1][0] / 25.4 * 72.0, _fallback[1][1] / 25.4 * 72.0)

    return size



# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from sdaps import model
from sdaps import script

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


parser = script.add_project_subparser("recognize",
    help=_("Run the optical mark recognition."),
    description=_("""Iterates over all images and runs the optical mark
    recognition. It will reevaluate sheets even if "recognize" has already
    run or manual changes were made."""))

parser.add_argument('--identify',
    help=_("Only identify the page properties, but don't recognize the checkbox states."),
    action="store_true",
    default=False)

parser.add_argument('--rerun', '-r',
    help=_("Rerun the recognition for all pages. The default is to skip all pages that were recognized or verified already."),
    action="store_true",
    default=False)

@script.connect(parser)
@script.logfile
def recognize(cmdline):
    survey = model.survey.Survey.load(cmdline['project'])
    from sdaps import recognize

    if not cmdline['rerun']:
        filter = lambda : not (survey.sheet.verified or survey.sheet.recognized)
    else:
        filter = lambda: True

    if cmdline['identify']:
        return recognize.identify(survey, filter)
    else:
        return recognize.recognize(survey, filter)





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module implements a simple export/import to/from CSV files.
"""

import os
import sys

from sdaps import model
from sdaps import script

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


parser = script.add_subparser("csv",
    help=_("Import or export data to/from CSV files."),
    description=_("""Import or export data to/from a CSV file. The first line
    is a header which defines questionnaire_id and global_id, and a column
    for each checkbox and textfield. Note that the import is currently very
    limited, as you need to specifiy the questionnaire ID to select the sheet
    which should be updated."""))

# Set required as an attribute rather than kwarg so that it works with python <3.7
subparser = parser.add_subparsers(dest='subcommand')
subparser.required = True

export = subparser.add_parser('export',
    help=_("Export data to CSV file."))
script.add_project_argument(export)

export.add_argument('-o', '--output',
    help=_("Filename to store the data to (default: data_%%i.csv)"))
export.add_argument('-d', '--delimiter',
    help=_("The delimiter used in the CSV file (default ',')"),
    default=',',
    action='store')
export.add_argument('-f', '--filter',
    help=_("Filter to only export a partial dataset."))
export.add_argument('--images',
    help=_("Export images of freeform fields."),
    dest='export_images',
    action='store_const',
    const=True,
    default=False)
export.add_argument('--question-images',
    help=_("Export an image for each question that includes all boxes."),
    dest='export_question_images',
    action='store_const',
    const=True,
    default=False)
export.add_argument('--quality',
    help=_("Export the recognition quality for each checkbox."),
    dest='export_quality',
    action='store_const',
    const=True,
    default=False)
export.set_defaults(direction='export')

import_ = subparser.add_parser('import',
    help=_("Import data to from a CSV file."))
script.add_project_argument(import_)

import_.add_argument('file',
    help=_("The file to import."))
import_.set_defaults(direction='import')

@script.connect(parser)
@script.logfile
def csvdata(cmdline):
    from sdaps import csvdata
    from sdaps.utils.image import ImageWriter

    survey = model.survey.Survey.load(cmdline['project'])

    if cmdline['direction'] == 'export':
        if cmdline['output']:
            if cmdline['output'] == '-':
                outfd = os.dup(sys.stdout.fileno())
                outfile = os.fdopen(outfd, 'w')
            else:
                filename = cmdline['output']
                outfile = open(filename, 'w')
        else:
            filename = survey.new_path('data_%i.csv')
            outfile = open(filename, 'w')

        csvoptions = { 'delimiter' : cmdline['delimiter'] }

        if cmdline['export_images'] or cmdline['export_question_images'] and cmdline['output'] != '-':
            img_path = os.path.dirname(filename)
            img_prefix = os.path.join(os.path.splitext(os.path.basename(filename))[0], 'img')

            image_writer = ImageWriter(img_path, img_prefix)
        else:
            image_writer = None


        return csvdata.csvdata_export(survey, outfile, image_writer,
            filter=cmdline['filter'],
            export_images=cmdline['export_images'],
            export_question_images=cmdline['export_question_images'],
            export_quality=cmdline['export_quality'],
            csvoptions=csvoptions)
    elif cmdline['direction'] == 'import':
        return csvdata.csvdata_import(survey, open(cmdline['file'], 'r'))
    else:
        raise AssertionError





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
from gi.repository import Pango
from gi.repository import PangoCairo

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

import copy
from . import buddies


def paint_box(cr, mm_to_pt, x, y, box, key):
    cr.save()
    cr.set_matrix(mm_to_pt)
    cr.translate(x, y)

    cr.scale(25.4 / 300.0, 25.4 / 300.0)

    cr.set_source_rgb(0, 0, 0)
    cr.mask_surface(box[2][0], 0, 0)

    if box[3] and key in box[3] and box[3][key] is not None:
        cr.set_source_surface(box[3][key][0], box[3][key][1] - box[2][1], box[3][key][2] - box[2][2])
        cr.paint()

    cr.restore()

    cr.save()

    tmp_x, tmp_y = mm_to_pt.transform_point(x, y)
    tmp_width, tmp_height = mm_to_pt.transform_distance(8, 13)

    tmp_x, tmp_y = mm_to_pt.transform_point(x, y + 8.0)
    # Print bold if detected ON
    value = box[1][key] if key in box[1] else -1
    if box[0]:
        t = "<b>%.2f</b>" % value
    else:
        t = "%.2f" % value

    cr.move_to(tmp_x, tmp_y)

    layout = PangoCairo.create_layout(cr)
    layout.set_markup(t, -1)
    font = Pango.FontDescription("serif 6")
    layout.set_font_description(font)

    PangoCairo.show_layout(cr, layout)

    cr.restore()


def fill_page(cr, mm_to_pt, checkboxes, key):
    y = 15
    y_step = 13
    x_step = 10
    x_max = 210 - 15 - 8
    y_max = 297 - 15 - 8

    while y < y_max:
        x = 15
        while x < x_max:

            if len(checkboxes) == 0:
                return
            box = checkboxes.pop(0)

            paint_box(cr, mm_to_pt, x, y, box, key)

            x += x_step
        y += y_step


def boxgallery(survey, debugrecognition):
    # Enable debug image creation in the C module
    if debugrecognition:
        from sdaps import image
        image.enable_debug_surface_creation(True)

    survey.questionnaire.boxgallery.init(debugrecognition)
    survey.iterate_progressbar(survey.questionnaire.boxgallery.get_checkbox_images)
    checkboxes = survey.questionnaire.boxgallery.checkboxes
    survey.questionnaire.boxgallery.clean()

    keys = set()
    for checkbox in checkboxes:
        keys = keys.union(iter(checkbox[1].keys()))

    for key in keys:
        print(_("Rendering boxgallery for metric \"%s\"." % key))
        draw_list = copy.copy(checkboxes)
        draw_list.sort(key=lambda x: x[1][key] if key in x[1] else 0)

        # Hardcode 300dpi
        # Hardcode the mm size:
        # 3.5 + 0.4mm = 3.9mm
        mm_to_pt = cairo.Matrix(72.0 / 25.4, 0, 0, 72.0 / 25.4, 0, 0)

        page = 1
        pdf = cairo.PDFSurface(survey.path('boxgallery-%s.pdf' % key), 595, 842)
        cr = cairo.Context(pdf)

        while len(draw_list) > 0:
            fill_page(cr, mm_to_pt, draw_list, key)
            cr.show_page()
            pdf.flush()

            page += 1

        del pdf
        del cr




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''data model for sdaps

Survey
======

survey ::

    +-- questionnaire
    |        +-- qobjects ---------------+
    |                +-- boxes ----------+
    +-- sheets                           |
            +-- data dictionary          |
            |        +-- data objects <--+
            +-- images


The questionnaire tree resembles the structer of the questionnaire and the
internal structure of the programm. It is set up by the setup script and not
altered(much) later.

The sheets tree holds the data of the answered questionnaires. The sheet
objects are created while the scanned images are added to the project.

Iterating through sheets and jumping
====================================

Survey provides a function iterate. It iterates through the sheets and calls a
function for each. Inside this function, the sheet is accessible through
survey.sheet(a property, pointing to the actual sheet)
The functions goto_something jump the survey.sheet pointer to the specified
sheet.

Buddies
=======

A script may define a buddy class for a class in the model. Then each object of
that model class is accompanied by an object of the buddy class.
The buddy object is instanciated automatically upon first access and then
cached. It is not saved on disc.

The buddy system allows you to define additional methods and attributes for
the objects in the model.

'''

from . import buddy
from . import data
from . import questionnaire
from . import sheet
from . import survey





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
from gi.repository import GLib, GObject
import cairo

from sdaps import model
from sdaps import defs

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

def markup_escape_text(text):
    # Unfortunately the API returns a byte string, so we need to decode it
    # as the formatting would not work otherwise.
    return GLib.markup_escape_text(text)

class Questionnaire(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'widget'
    obj_class = model.questionnaire.Questionnaire

    def __init__(self, *args):
        model.buddy.Buddy.__init__(self, *args)

        self.obj.connect_data_changed(self.data_changed)

        self._current_image = None
        self._notify_ensure_visible = list()

    def data_changed(self, questionnaire, qobj, obj, name, old_value):
        # Simply sync everything on every change.
        self.sync_state()

    def create_widget(self):
        self.box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        # First some global options
        widget = Gtk.Label()
        widget.set_markup(_('<b>Global Properties</b>'))
        widget.props.xalign = 0.0
        self.box.pack_start(widget, False, True, 0)

        self.valid_checkbox = Gtk.CheckButton.new_with_label(_('Sheet valid'))
        self.sheet_verified_checkbox = Gtk.CheckButton.new_with_label(_('Sheet Verified'))
        self.page_verified_checkbox = Gtk.CheckButton.new_with_label(_('Page Verified'))
        self.empty_checkbox = Gtk.CheckButton.new_with_label(_('Empty'))
        self.empty_checkbox.set_sensitive(False)

        frame = Gtk.Frame()
        self.review_textbox = Gtk.TextView()
        self.review_textbox.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        self.review_buffer = self.review_textbox.get_buffer()
        self.review_buffer.connect('changed', self.review_buffer_changed_cb)
        frame.add(self.review_textbox)

        self.review_comments = Gtk.CheckButton.new_with_label(_('Empty'))

        self.valid_checkbox.connect('toggled', self.toggled_valid_cb)
        self.page_verified_checkbox.connect('toggled', self.toggled_verified_cb)

        indent = Gtk.Alignment()
        indent.set_padding(0, 0, 10, 0)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.qid = Gtk.Label()
        self.qid.set_markup(_('<b>Questionnaire ID: </b>') + markup_escape_text(str(self.obj.survey.sheet.questionnaire_id)))
        self.qid.props.xalign = 0.0

        indent.add(vbox)

        vbox.add(self.qid)
        vbox.add(self.valid_checkbox)
        vbox.add(self.sheet_verified_checkbox)
        vbox.add(self.page_verified_checkbox)
        vbox.add(self.empty_checkbox)
        vbox.add(frame)

        self.box.pack_start(indent, False, True, 0)

        # And all the questions
        for qobject in self.obj.qobjects:
            widget = qobject.widget.create_widget()
            if widget is not None:
                self.box.pack_start(widget, False, True, 0)

        return self.box

    def sync_state(self, image=None):
        # This is to keep track of the currently active image.
        if image:
            self._current_image = image
            assert self._current_image in self.obj.survey.sheet.images

        for qobject in self.obj.qobjects:
            qobject.widget.sync_state()

        self.qid.set_markup(_('<b>Questionnaire ID: </b>') + markup_escape_text(str(self.obj.survey.sheet.questionnaire_id)))
        self.valid_checkbox.set_active(self.obj.survey.sheet.valid)
        self.sheet_verified_checkbox.set_active(self.obj.survey.sheet.verified)
        self.page_verified_checkbox.set_active(self._current_image.verified)
        self.empty_checkbox.set_active(self.obj.survey.sheet.empty)

        # Only update the text if it changed (or else recursion hits)
        start = self.review_buffer.get_start_iter()
        end = self.review_buffer.get_end_iter()
        currtext = self.review_buffer.get_text(start, end, False)
        review_comment = self.obj.survey.sheet.review_comment if self.obj.survey.sheet.review_comment else ''
        if review_comment != currtext:
            self.review_buffer.set_text(review_comment)

    def ensure_visible(self, widget):
        for func in self._notify_ensure_visible:
            func(widget)

    def connect_ensure_visible(self, func):
        self._notify_ensure_visible.append(func)

    def disconnect_ensure_visible(self, func):
        self._notify_ensure_visible.remove(func)

    def review_buffer_changed_cb(self, buf):
        start = self.review_buffer.get_start_iter()
        end = self.review_buffer.get_end_iter()
        currtext = self.review_buffer.get_text(start, end, False)

        self.obj.survey.sheet.review_comment = currtext

    def toggled_valid_cb(self, widget):
        self.obj.survey.sheet.valid = widget.get_active()

    def toggled_verified_cb(self, widget):
        self._current_image.verified = widget.get_active()

class QObject(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'widget'
    obj_class = model.questionnaire.QObject

    def create_widget(self):
        self.widget = None
        self.review_buffer = None

        return self.widget

    def make_heading(self, title):
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        vbox.set_margin_top(12)
        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        label = Gtk.Label()
        label.set_markup('<b>%s %s</b>' % (self.obj.id_str(), markup_escape_text(title)))
        label.set_halign(Gtk.Align.START)
        label.show()
        hbox.add(label)

        image = Gtk.Image()
        image.set_from_icon_name('document-edit-symbolic', Gtk.IconSize.BUTTON)
        toggle = Gtk.ToggleButton(image=image)
        toggle.set_halign(Gtk.Align.END)
        toggle.set_margin_start(12)
        hbox.add(toggle)
        image.connect("button-press-event", lambda *args: print(pressed))
        vbox.add(hbox)

        self.review_frame = Gtk.Frame()
        self.review_frame.set_margin_start(12)
        self.review_textbox = Gtk.TextView()
        self.review_textbox.show()
        self.review_textbox.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        self.review_buffer = self.review_textbox.get_buffer()
        self.review_buffer.connect('changed', self.review_buffer_changed_cb)
        self.review_frame.add(self.review_textbox)
        self.review_frame.hide()
        self.review_frame.set_no_show_all(True)
        vbox.add(self.review_frame)

        toggle.bind_property("active",
                             self.review_frame, "visible",
                             GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL)

        return vbox

    def sync_state(self):
        for box in self.obj.boxes:
            box.widget.sync_state()

        # Only update the text if it changed (or else recursion hits)
        if self.review_buffer:
            start = self.review_buffer.get_start_iter()
            end = self.review_buffer.get_end_iter()
            currtext = self.review_buffer.get_text(start, end, False)
            review_comment = self.obj.data.review_comment if self.obj.data.review_comment else ''
            if review_comment != currtext:
                self.review_frame.set_visible(bool(review_comment))
                self.review_buffer.set_text(review_comment)

    def focus(self):
        self.obj.question.questionnaire.widget.ensure_visible(self.widget)

        if len(boxes) > 0:
            self.obj.boxes[0].widget.focus()

    def review_buffer_changed_cb(self, buf):
        start = self.review_buffer.get_start_iter()
        end = self.review_buffer.get_end_iter()
        currtext = self.review_buffer.get_text(start, end, False)

        self.obj.data.review_comment = currtext


class Head(QObject, metaclass=model.buddy.Register):

    name = 'widget'
    obj_class = model.questionnaire.Head

    def create_widget(self):
        self.widget = self.make_heading(self.obj.title)

        return self.widget


class Question(QObject, metaclass=model.buddy.Register):

    name = 'widget'
    obj_class = model.questionnaire.Question

    def create_widget(self):
        self.widget = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.label = self.make_heading(self.obj.question)

        self.widget.pack_start(self.label, False, True, 0)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        vbox.set_margin_start(12)
        self.widget.pack_end(vbox, False, True, 0)

        for box in self.obj.boxes:
            widget = box.widget.create_widget()
            if widget is not None:
                vbox.pack_start(widget, False, True, 0)

        return self.widget

class Range(Question, metaclass=model.buddy.Register):

    name = 'widget'
    obj_class = model.questionnaire.Range

    def create_widget(self):
        self.widget = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.label = self.make_heading(self.obj.question)

        self.widget.pack_start(self.label, False, True, 0)

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        lower_label = Gtk.Label()
        lower_label.set_text(self.obj.answers[0])
        lower_label.props.xalign = 1.0

        upper_label = Gtk.Label()
        upper_label.set_text(self.obj.answers[1])
        upper_label.props.xalign = 0.0

        sizegroup = Gtk.SizeGroup()
        sizegroup.set_mode(Gtk.SizeGroupMode.BOTH)
        sizegroup.add_widget(lower_label)
        sizegroup.add_widget(upper_label)

        hbox.pack_start(lower_label, True, True, 0)

        # Maybe use radiobuttons instead?
        for box in self.obj.boxes:
            widget = box.widget.create_widget()
            if widget is not None:
                hbox.pack_start(widget, False, True, 0)

        hbox.pack_start(upper_label, True, True, 0)

        self.widget.pack_start(hbox, False, True, 0)

        return self.widget


class Box(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'widget'
    obj_class = model.questionnaire.Checkbox

    def create_widget(self):
        if self.obj.text:
            self.widget = Gtk.CheckButton.new_with_label(self.obj.text)
        else:
            self.widget = Gtk.CheckButton.new_with_label('')
        self.widget.connect('toggled', self.toggled_cb)

        return self.widget

    def sync_state(self):
        self.widget.props.active = self.obj.data.state

    def toggled_cb(self, widget):
        self.obj.data.state = widget.props.active

    def focus(self):
        self.widget.grab_focus()

        self.obj.question.questionnaire.widget.ensure_visible(self.obj.question.widget.widget)
        self.obj.question.questionnaire.widget.ensure_visible(self.widget)

class Checkbox(Box, metaclass=model.buddy.Register):

    name = 'widget'
    obj_class = model.questionnaire.Checkbox


class Textbox(Box, metaclass=model.buddy.Register):

    name = 'widget'
    obj_class = model.questionnaire.Textbox

    def create_widget(self):
        self.widget = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.checkbox = Gtk.CheckButton.new_with_label(self.obj.text)
        self.checkbox.connect('toggled', self.toggled_cb)

        self.widget.add(self.checkbox)

        frame = Gtk.Frame()
        frame.set_margin_start(12)
        self.widget.pack_end(frame, False, True, 0)

        self.textbox = Gtk.TextView()
        self.textbox.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        self.buffer = self.textbox.get_buffer()
        self.buffer.connect('changed', self.buffer_changed_cb)

        frame.add(self.textbox)

        return self.widget

    def buffer_changed_cb(self, buf):
        start = buf.get_start_iter()
        end = buf.get_end_iter()
        self.obj.data.text = buf.get_text(start, end, False)

    def sync_state(self):
        self.checkbox.props.active = self.obj.data.state

        self.textbox.props.sensitive = self.obj.data.state

        # Only update the text if it changed (or else recursion hits)
        start = self.buffer.get_start_iter()
        end = self.buffer.get_end_iter()
        currtext = self.buffer.get_text(start, end, False)
        if self.obj.data.text != currtext:
            self.buffer.set_text(self.obj.data.text)

    def focus(self):
        if self.textbox.props.sensitive:
            self.textbox.grab_focus()
        else:
            self.checkbox.grab_focus()

        self.obj.question.questionnaire.widget.ensure_visible(self.obj.question.widget.widget)
        self.obj.question.questionnaire.widget.ensure_visible(self.widget)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2007-2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2007-2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This modules contains a GTK+ user interface. With the help of this GUI it is
possible to manually correct the automatic recognition and do basic debugging.
"""

import gi
gi.require_version('Gtk', '3.0')

from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Gdk
import os
import time
import sys
import signal

from sdaps import model
from sdaps import surface
from sdaps import clifilter
from sdaps import defs
from sdaps import paths
from sdaps import log

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

from .sheet_widget import SheetWidget
from . import buddies
from . import widget_buddies


zoom_steps = [0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
              1.25, 1.5, 2.0, 2.5, 3.0]


def gui(survey, cmdline):
    filter = clifilter.clifilter(survey, cmdline['filter'])
    provider = Provider(survey, filter)
    if not provider.images:
        log.error(_("The survey does not have any images! Please add images (and run recognize) before using the GUI."))
        return 1

    try:
        # Exit the mainloop if Ctrl+C is pressed in the terminal.
        GLib.unix_signal_add_full(GLib.PRIORITY_HIGH, signal.SIGINT, lambda *args : Gtk.main_quit(), None)
    except AttributeError:
        # Whatever, it is only to enable Ctrl+C anyways
        pass

    MainWindow(provider).run()


class Provider(object):

    def __init__(self, survey, filter, by_quality=False):
        self._by_quality = by_quality

        self.survey = survey
        self.images = list()
        self.qualities = list()
        self.survey.iterate(self, filter)
        self.qualities.sort(reverse=False)
        self.index = 0

        # There may be no images. This error is
        # caught and printed in the "gui" function.
        if not self.images:
            return

        self.image.surface.load_rgb()
        self.survey.goto_sheet(self.image.sheet)
        #self._surface = None

    def __call__(self):
        # Add all images that are "valid" ie. everything except back side of
        # a simplex printout.
        new_images = [img for img in self.survey.sheet.images if not img.ignored]

        # Try to sort the images, which may fail if the page numbers have not been
        # recognised
        try:
            new_images.sort(key=lambda Image: Image.page_number)
        except TypeError:
            pass

        self.images.extend(new_images)
        # Insert each image of the sheet into the qualities array
        for i in range(len(new_images)):
            self.qualities.append((self.survey.sheet.quality, len(self.qualities)))

    def next(self, cycle=True):
        if self.index >= len(self.images) - 1:
            if cycle:
                new_index = 0
            else:
                return False
        else:
            new_index = self.index + 1

        self.image.surface.clean()

        self.index = new_index

        self.image.surface.load_rgb()
        self.survey.goto_sheet(self.image.sheet)

        return True

    def previous(self, cycle=True):
        if self.index <= 0:
            if cycle:
                new_index = len(self.images) - 1
            else:
                return False
        else:
            new_index = self.index - 1

        self.image.surface.clean()

        self.index = new_index

        self.image.surface.load_rgb()
        self.survey.goto_sheet(self.image.sheet)

        return True

    def goto(self, index):
        if index >= 0 and index < len(self.images):
            self.image.surface.clean()
            self.index = index
            self.image.surface.load_rgb()
            self.survey.goto_sheet(self.image.sheet)

    def set_sort_by_quality(self, value):
        self.image.surface.clean()
        self._by_quality = value
        self.image.surface.load_rgb()
        self.survey.goto_sheet(self.image.sheet)

    def get_image(self):
        if self._by_quality:
            return self.images[self.qualities[self.index][1]]
        else:
            return self.images[self.index]

    image = property(get_image)


class MainWindow(object):

    def __init__(self, provider):
        self.about_dialog = None
        self.close_dialog = None
        self.ask_open_dialog = None
        self.provider = provider

        self._load_image = 0
        self._builder = Gtk.Builder()
        if paths.local_run:
            self._builder.add_from_file(
                os.path.join(os.path.dirname(__file__), 'main_window.ui'))
        else:
            self._builder.add_from_file(
                os.path.join(
                    paths.prefix,
                    'share', 'sdaps', 'ui', 'main_window.ui'))

        self._window = self._builder.get_object("main_window")
        self._builder.connect_signals(self)
        self._window.maximize()

        scrolled_window = self._builder.get_object("sheet_scrolled_window")
        self.sheet = SheetWidget(self.provider)
        self.sheet.show()
        scrolled_window.add(self.sheet)

        self.data_viewport = self._builder.get_object("data_view")
        widgets = provider.survey.questionnaire.widget.create_widget()
        widgets.show_all()
        self.data_viewport.add(widgets)

        provider.survey.questionnaire.widget.connect_ensure_visible(self.data_view_ensure_visible)

        self.sheet.connect('key-press-event', self.sheet_view_key_press)

        combo = self._builder.get_object("page_number_combo")
        cell = Gtk.CellRendererText()
        combo.pack_start(cell, True)
        combo.add_attribute(cell, 'text', 0)

        store = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_INT)
        store.append(row=(_("Page|Invalid"), -1))
        for i in range(self.provider.survey.questionnaire.page_count):
            store.append(row=(
                ungettext("Page %i", "Page %i", i + 1) % (i + 1), i + 1))

        combo.set_model(store)

        self.sheet.props.zoom = 0.2

        # So the buttons are insensitive
        self.update_ui()

    def zoom_in(self, *args):
        cur_zoom = self.sheet.props.zoom
        try:
            i = zoom_steps.index(cur_zoom)
            i += 1
            if i < len(zoom_steps):
                self.sheet.props.zoom = zoom_steps[i]
        except:
            self.sheet.props.zoom = 1.0

    def zoom_out(self, *args):
        cur_zoom = self.sheet.props.zoom
        try:
            i = zoom_steps.index(cur_zoom)
            i -= 1
            if i >= 0:
                self.sheet.props.zoom = zoom_steps[i]
        except:
            self.sheet.props.zoom = 1.0

    def null_event_handler(self, *args):
        return True

    def show_about_dialog(self, *args):
        if not self.about_dialog:
            self.about_dialog = Gtk.AboutDialog()
            self.about_dialog.set_program_name("SDAPS")
            #self.about_dialog.set_version("")
            self.about_dialog.set_authors(
                ["Benjamin Berg <benjamin@sipsolution.net>",
                 "Ferdinand Schwenk <ferdisdot@gmail.com>",
                 "Christoph Simon <post@christoph-simon.eu>",
                 "Tobias Simon <tobsimon@googlemail.com>"])
            self.about_dialog.set_copyright(_("Copyright © 2007-2014 The SDAPS Authors"))
            self.about_dialog.set_license_type(Gtk.License.GPL_3_0)
            self.about_dialog.set_comments(_("Scripts for data acquisition with paper based surveys"))
            self.about_dialog.set_website(_("http://sdaps.org"))
            self.about_dialog.set_translator_credits(_("translator-credits"))
            self.about_dialog.set_default_response(Gtk.ResponseType.CANCEL)

        self.about_dialog.set_transient_for(self._window)
        self.about_dialog.run()
        self.about_dialog.hide()

        return True

    def update_page_status(self):
        combo = self._builder.get_object("page_number_combo")
        turned_toggle = self._builder.get_object("turned_toggle")

        # Update the combobox
        if self.provider.image.survey_id == self.provider.survey.survey_id:
            page_number = self.provider.image.page_number
        else:
            page_number = -1

        # Find the page_number in the model
        model = combo.get_model()
        iter = model.get_iter_first()

        while iter:
            i = combo.get_model().get(iter, 1)[0]
            if page_number == i:
                combo.set_active_iter(iter)
                iter = None
            else:
                iter = model.iter_next(iter)

        # Update the toggle
        turned_toggle.set_active(self.provider.image.rotated or False)

    def update_ui(self):
        # Update the next/prev button states
        #next_button = self._builder.get_object("forward_toolbutton")
        #prev_button = self._builder.get_object("backward_toolbutton")

        #next_button.set_sensitive(True)
        #prev_button.set_sensitive(True)

        position_label = self._builder.get_object("position_label")
        quality_label = self._builder.get_object("quality_label")
        page_spin = self._builder.get_object("page_spin")
        position_label.set_text(_(" of %i") % len(self.provider.images))
        quality_label.set_text(_("Recognition Quality: %.2f") % self.provider.image.sheet.quality)
        #position_label.props.sensitive = True
        page_spin.set_range(1, len(self.provider.images))
        page_spin.set_value(self.provider.index + 1)

        self.update_page_status()
        self.sheet.update_state()

        self.provider.survey.questionnaire.widget.sync_state(self.provider.image)

    def go_to_previous_page(self, *args):
        if not self.provider.previous(cycle=False):
            dialog = Gtk.MessageDialog(
                flags=Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
                type=Gtk.MessageType.INFO,
                buttons=Gtk.ButtonsType.CANCEL,
                message_format=_("You have reached the first page of the survey. Would you like to go to the last page?"))
            dialog.set_transient_for(self._window)
            dialog.add_button(_("Go to last page"), Gtk.ResponseType.OK)
            if dialog.run() == Gtk.ResponseType.OK:
                self.provider.previous(cycle=True)
            dialog.destroy()
        self.update_ui()
        return True

    def go_to_page(self, page):
        if page == self.provider.index:
            return True

        self.provider.goto(int(page))

        self.update_ui()
        return True

    def go_to_next_page(self, *args):
        if not self.provider.next(cycle=False):
            dialog = Gtk.MessageDialog(
                flags=Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
                type=Gtk.MessageType.INFO,
                buttons=Gtk.ButtonsType.CANCEL,
                message_format=_("You have reached the last page of the survey. Would you like to go to the first page?"))
            dialog.set_transient_for(self._window)
            dialog.add_button(_("Go to first page"), Gtk.ResponseType.OK)
            if dialog.run() == Gtk.ResponseType.OK:
                self.provider.next(cycle=True)
            dialog.destroy()

        self.update_ui()
        return True

    def page_spin_value_changed_cb(self, *args):
        page_spin = self._builder.get_object("page_spin")
        page = page_spin.get_value() - 1
        self.go_to_page(page)

    def page_number_combo_changed_cb(self, *args):
        combo = self._builder.get_object("page_number_combo")
        active = combo.get_active_iter()
        page_number = combo.get_model().get(active, 1)[0]
        if self.provider.image.page_number != page_number:
            if page_number != -1:
                self.provider.image.page_number = page_number
                self.provider.image.survey_id = self.provider.survey.survey_id
            else:
                self.provider.image.page_number = None
                self.provider.image.survey_id = None
            self.update_ui()
            return False

    def turned_toggle_toggled_cb(self, *args):
        toggle = self._builder.get_object("turned_toggle")
        rotated = toggle.get_active()
        if self.provider.image.rotated != rotated:
            self.provider.image.rotated = rotated
            self.provider.image.surface.load_rgb()
            self.update_ui()
        return False

    def quality_sort_toggle_toggled_cb(self, *args):
        toggle = self._builder.get_object("quality_sort_toggle")
        self.provider.set_sort_by_quality(toggle.get_active())
        self.update_ui()
        return False

    def toggle_fullscreen(self, *args):
        flags = self._window.get_window().get_state()
        if flags & Gdk.WindowState.FULLSCREEN:
            self._window.unfullscreen()
        else:
            self._window.fullscreen()
        return True

    def save_project(self, *args):
        self.provider.survey.save()
        return True

    def data_view_ensure_visible(self, widget):
        allocation = widget.get_allocation()

        vadj = self.data_viewport.props.vadjustment
        lower = vadj.props.lower
        upper = vadj.props.upper
        value = vadj.props.value
        page_size = vadj.props.page_size

        value = max(value, allocation.y + allocation.height - page_size)
        value = min(value, allocation.y)
        value = max(value, lower)
        value = min(value, upper)

        vadj.props.value = value

    def sheet_view_key_press(self, window, event):
        # Go to the next when Enter or Tab is pressed
        enter_keyvals = [Gdk.keyval_from_name(k) for k in ["Return", "KP_Enter", "ISO_Enter"]]
        if event.keyval in enter_keyvals:
            # If "Return" is pressed, then the examiner figured that the data
            # is good.

            # Mark as verified
            self.provider.image.verified = True

            # Mark the sheet as valid if it is verified
            if self.provider.image.sheet.verified:
                self.provider.image.sheet.valid = True

            if event.state & Gdk.ModifierType.SHIFT_MASK:
                self.go_to_previous_page()
            else:
                self.go_to_next_page()
            return True
        elif event.keyval == Gdk.KEY_Tab or event.keyval == Gdk.KEY_KP_Tab or event.keyval == Gdk.KEY_ISO_Left_Tab:
            # Allow tabbing out with Ctrl
            if event.state & Gdk.ModifierType.CONTROL_MASK:
                return False

            if event.state & Gdk.ModifierType.SHIFT_MASK:
                self.go_to_previous_page()
            else:
                self.go_to_next_page()
            return True

        return False

    def quit_application(self, *args):
        if not self.close_dialog:
            self.close_dialog = Gtk.MessageDialog(
                parent=self._window,
                flags=Gtk.DialogFlags.MODAL,
                type=Gtk.MessageType.WARNING)
            self.close_dialog.add_buttons(
                _("Close without saving"), Gtk.ResponseType.CLOSE,
                Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                Gtk.STOCK_SAVE, Gtk.ResponseType.OK)
            self.close_dialog.set_markup(
                _("<b>Save the project before closing?</b>\n\nIf you do not save you may loose data."))
            self.close_dialog.set_default_response(Gtk.ResponseType.CANCEL)

        response = self.close_dialog.run()
        self.close_dialog.hide()

        if response == Gtk.ResponseType.CLOSE:
            Gtk.main_quit()
            return False
        elif response == Gtk.ResponseType.OK:
            self.save_project()
            Gtk.main_quit()
            return False
        else:
            return True

    def run(self):
        self._window.show()
        Gtk.main()


